"""Griffe TypingDoc package.

Griffe extension for PEP 727 - Documentation Metadata in Typing.
"""

from __future__ import annotations

from griffe_typingdoc._extension import TypingDocExtension

__all__: list[str] = ["TypingDocExtension"]
